@@ -1,6 +1,6 @@
 /************************************************************************
 **
-**  Copyright (C) 2016-2022 Kevin B. Hendricks, Stratford, Ontario, Canada
+**  Copyright (C) 2016-2024 Kevin B. Hendricks, Stratford, Ontario, Canada
 **  Copyright (C) 2012      John Schember <john@nachtimwald.com>
 **  Copyright (C) 2009-2011 Strahinja Markovic  <strahinja.markovic@gmail.com>
 **
@@ -49,6 +49,7 @@
 #include <QDate>
 #include <QTime>
 #include <QUrl>
+#include <QStringDecoder>
 #include <QDebug>
 
 #include "BookManipulation/FolderKeeper.h"
@@ -57,7 +58,6 @@
 #include "Misc/MediaTypes.h"
 #include "Misc/FontObfuscation.h"
 #include "Misc/HTMLEncodingResolver.h"
-#include "Misc/QCodePage437Codec.h"
 #include "Misc/SettingsStore.h"
 #include "Misc/Utility.h"
 #include "ResourceObjects/CSSResource.h"
@@ -90,7 +90,9 @@ static const QString CONTAINER_XML       = "<?xml version=\"1.0\" encoding=\"UTF
         "   </rootfiles>\n"
         "</container>\n";
 
-static QCodePage437Codec *cp437 = 0;
+static QStringDecoder *cp437 = nullptr;
+
+static  const QString SEP = QString(QChar(31));
 
 // Constructor;
 // The parameter is the file to be imported
@@ -151,9 +153,9 @@ QSharedPointer<Book> ImportEPUB::GetBook(bool extract_metadata)
     }
     
     if (!notInManifest.isEmpty()) {
-        QApplication::restoreOverrideCursor();
-        Utility::DisplayStdWarningDialog(tr("Files exist in epub that are not listed in the manifest, they will be ignored"), notInManifest.join("\n"));
-        QApplication::setOverrideCursor(Qt::WaitCursor);
+        QString warning = tr("Files exist in epub that are not listed in the manifest, they will be ignored.");
+        warning = warning + SEP + notInManifest.join("\n"); 
+        AddLoadWarning(warning);
     }
 
     LoadFolderStructure();
@@ -172,7 +174,8 @@ QSharedPointer<Book> ImportEPUB::GetBook(bool extract_metadata)
         }
     }
 
-    bool checkit = ((ss.cleanOn() & CLEANON_OPEN) == CLEANON_OPEN);
+    bool autofix = ((ss.cleanOn() & CLEANON_OPEN) == CLEANON_OPEN);
+    bool checkit = true;
 
     QFuture<std::pair<HTMLResource*, bool> > html_future;
     html_future = QtConcurrent::mapped(hresources, std::bind(InitialLoadAndCheckOneHTMLFile, std::placeholders::_1, checkit));
@@ -183,23 +186,28 @@ QSharedPointer<Book> ImportEPUB::GetBook(bool extract_metadata)
         }
     }
     if (!non_well_formed.isEmpty()) {
-        QApplication::restoreOverrideCursor();
-        if (QMessageBox::Yes == QMessageBox::warning(QApplication::activeWindow(),
-                tr("Sigil"),
-                tr("This EPUB has HTML files that are not well formed or are "
-                   "missing a DOCTYPE, html, head or body elements. "
-                   "Sigil can automatically fix these files, although "
-                   "this may very rarely result in minor data loss in extreme circumstances.\n\n"
-                   "Do you want to automatically fix the files?"),
-                QMessageBox::Yes|QMessageBox::No)) 
-        {
+        QStringList problem_files;
+        foreach(HTMLResource* htmlres, non_well_formed) {
+            problem_files << htmlres->GetRelativePath();
+        }
+        if (autofix) {
             foreach(HTMLResource* htmlres, non_well_formed) {
                 QString fixed_text = CleanSource::Mend(htmlres->GetText(),htmlres->GetEpubVersion());
                 htmlres->SetText(fixed_text);
             }
             non_well_formed.clear();
+            QString warning = tr("This EPUB had HTML files that were not well formed or are "
+                   "missing a DOCTYPE, html, head or body elements.<br/><br>They were automatically fixed based on "
+                   "your Preference setting to Clean on Open.");
+            warning = warning + SEP + problem_files.join("\n");
+            AddLoadWarning(warning);
+        } else {
+            QString warning = tr("This EPUB has HTML files that are not well formed or are "
+                   "missing a DOCTYPE, html, head or body elements.<br/></br>Fix these manually or use "
+                                 "Sigil's Mend tool to automatically fixed these errors or omissions.");
+            warning = warning + SEP + problem_files.join("\n");
+            AddLoadWarning(warning);
         }
-        QApplication::setOverrideCursor(Qt::WaitCursor);
     }
 
     ProcessFontFiles(resources, encrypted_files);
@@ -266,13 +274,14 @@ QSharedPointer<Book> ImportEPUB::GetBook(bool extract_metadata)
 
 QHash<QString, QString> ImportEPUB::ParseEncryptionXml()
 {
-    QString encrpytion_xml_path = m_ExtractedFolderPath + "/META-INF/encryption.xml";
+    QString encryption_xml_path = m_ExtractedFolderPath + "/META-INF/encryption.xml";
 
-    if (!QFileInfo(encrpytion_xml_path).exists()) {
+    if (!QFileInfo(encryption_xml_path).exists()) {
         return QHash<QString, QString>();
     }
 
-    QXmlStreamReader encryption(Utility::ReadUnicodeTextFile(encrpytion_xml_path));
+    // store away the font obfuscation info
+    QXmlStreamReader encryption(Utility::ReadUnicodeTextFile(encryption_xml_path));
     QHash<QString, QString> encrypted_files;
     QString encryption_algo;
     QString uri;
@@ -302,6 +311,8 @@ QHash<QString, QString> ImportEPUB::ParseEncryptionXml()
         throw (EPUBLoadParseError(error.toStdString()));
     }
 
+    // remove the encryption.xml file, it will be re-created as needed
+    Utility::SDeleteFile(encryption_xml_path);
     return encrypted_files;
 }
 
@@ -402,7 +413,7 @@ void ImportEPUB::ExtractContainer()
 {
     int res = 0;
     if (!cp437) {
-        cp437 = new QCodePage437Codec();
+        cp437 = new QStringDecoder("IBM437");
     }
 #ifdef Q_OS_WIN32
     zlib_filefunc64_def ffunc;
@@ -416,6 +427,8 @@ void ImportEPUB::ExtractContainer()
         throw (EPUBLoadParseError(QString(QObject::tr("Cannot unzip EPUB: %1")).arg(QDir::toNativeSeparators(m_FullFilePath)).toStdString()));
     }
 
+    // Note: zip archives can do utf-8 but they do NOT have a standard for Unicode NormalizationForm
+    // we will choose to use NFC
     res = unzGoToFirstFile(zfile);
 
     if (res == UNZ_OK) {
@@ -427,10 +440,14 @@ void ImportEPUB::ExtractContainer()
             QString qfile_name;
             QString cp437_file_name;
             qfile_name = QString::fromUtf8(file_name);
+            // must do this unconditionally by epub spec
+            qfile_name = qfile_name.normalized(QString::NormalizationForm_C);
             if (!(file_info.flag & (1<<11))) {
                 // General purpose bit 11 says the filename is utf-8 encoded. If not set then
                 // IBM 437 encoding might be used.
-                cp437_file_name = cp437->toUnicode(file_name);
+                cp437_file_name = cp437->decode(file_name);
+                // must do this unconditionally by epub spec
+                cp437_file_name = cp437_file_name.normalized(QString::NormalizationForm_C);
             }
             QDate moddate = QDate(file_info.tmu_date.tm_year,
                                   file_info.tmu_date.tm_mon + 1,
@@ -626,7 +643,7 @@ void ImportEPUB::LocateOPF()
     }
 
     if (num_opf > 1) {
-        Utility::DisplayStdWarningDialog(tr("This epub has multiple renditions (multiple OPF files). Editing this epub in Sigil will produce a normal single rendition epub using only the main (first) OPF file found."),"");
+        AddLoadWarning(QObject::tr("This epub has multiple renditions (multiple OPF files). Editing this epub in Sigil will produce a normal single rendition epub using only the main (first) OPF file found."));
     }
 
     if (m_OPFFilePath.isEmpty() || !QFile::exists(m_OPFFilePath)) {
@@ -781,6 +798,8 @@ void ImportEPUB::ReadManifestItemElement(QXmlStreamReader *opf_reader)
     if (href.indexOf(':') == -1) {
         // we know we have a relative href to a file so no fragments can exist
         apath = Utility::URLDecodePath(href);
+        // must do this unconditionally by epub spec
+        apath = apath.normalized(QString::NormalizationForm_C);
     }
     // for hrefs pointing outside the epub, apath will be empty
     // qDebug() << "ImportEpub with Manifest item: " << href << apath;
@@ -788,11 +807,12 @@ void ImportEPUB::ReadManifestItemElement(QXmlStreamReader *opf_reader)
 
     // validate the media type if we can, and warn otherwise
     QString group = MediaTypes::instance()->GetGroupFromMediaType(type,"");
-    QString ext_mtype = MediaTypes::instance()->GetMediaTypeFromExtension(extension, "");
+    QString ext_mtype = MediaTypes::instance()->GetMediaTypeFromExtension(extension,"");
     if (type.isEmpty() || group.isEmpty()) {
         const QString load_warning = QObject::tr("The OPF uses an unrecognized media type \"%1\" for file \"%2\"").arg(type).arg(QFileInfo(apath).fileName()) +
             " - " + QObject::tr("A temporary media type of \"%1\" has been assigned. You should edit your OPF file to fix this problem.").arg(ext_mtype);
         AddLoadWarning(load_warning);
+        if (!ext_mtype.isEmpty()) type = ext_mtype;
     }
 
     if (!apath.isEmpty()) {
@@ -800,7 +820,7 @@ void ImportEPUB::ReadManifestItemElement(QXmlStreamReader *opf_reader)
         // find the epub root relative file path from the opf location and the item href
         QString file_path = m_opfDir.absolutePath() + "/" + apath;
         file_path = Utility::resolveRelativeSegmentsInFilePath(file_path,"/");
-        file_path = file_path.remove(0, m_ExtractedFolderPath.length() + 1); 
+        file_path = file_path.remove(0, m_ExtractedFolderPath.length() + 1);
     
         // Manifest Items may *NOT* live in the META-INF and the mimetype file should NOT be manifested
         if (file_path.startsWith("META-INF/") || (file_path == "mimetype")) {
@@ -858,24 +878,26 @@ void ImportEPUB::LocateOrCreateNCX(const QString &ncx_id_on_spine)
 
     // handle the normal/proper case of an ncx id on the spine matching an ncx candidate that exists
     if (!m_NCXId.isEmpty() && m_NcxCandidates.contains(m_NCXId)) {
-        QString bookpath;
         ncx_href = m_NcxCandidates[ m_NCXId ];
         m_NCXFilePath = QFileInfo(m_OPFFilePath).absolutePath() % "/" % ncx_href;
         m_NCXFilePath = Utility::resolveRelativeSegmentsInFilePath(m_NCXFilePath, "/");
-        bookpath = m_NCXFilePath.right(m_NCXFilePath.length() - m_ExtractedFolderPath.length() - 1);
-        QString ncx_text = CleanSource::ProcessXML(Utility::ReadUnicodeTextFile(m_NCXFilePath),
-                                                   "application/x-dtbncx+xml");
-        NCXResource* resource = m_Book->GetFolderKeeper()->AddNCXToFolder(m_PackageVersion, bookpath);
-        resource->SetText(ncx_text);
-        resource->SaveToDisk(false);
-        if (m_FileInfoFromZip.contains(bookpath)) {
-            std::tuple<size_t, QString, QString> ainfo = m_FileInfoFromZip[bookpath];
-            resource->SetSavedSize(std::get<0>(ainfo));
-            resource->SetSavedCRC32(std::get<1>(ainfo));
-            resource->SetSavedDate(std::get<2>(ainfo));
-	    }
-        m_NCXNotInManifest = false;
-        return;
+        if (QFile::exists(m_NCXFilePath)) {
+            QString ncx_text = Utility::ReadUnicodeTextFile(m_NCXFilePath);
+            ncx_text = CleanSource::ProcessXML(ncx_text, "application/x-dtbncx+xml");
+            QString bookpath = m_NCXFilePath.right(m_NCXFilePath.length() - m_ExtractedFolderPath.length() - 1);
+            NCXResource* resource = m_Book->GetFolderKeeper()->AddNCXToFolder(m_PackageVersion, bookpath);
+            resource->SetText(ncx_text);
+            resource->SaveToDisk(false);
+            if (m_FileInfoFromZip.contains(bookpath)) {
+                std::tuple<size_t, QString, QString> ainfo = m_FileInfoFromZip[bookpath];
+                resource->SetSavedSize(std::get<0>(ainfo));
+                resource->SetSavedCRC32(std::get<1>(ainfo));
+                resource->SetSavedDate(std::get<2>(ainfo));
+	        }
+            m_NCXNotInManifest = false;
+            return;
+        }
+        // No NCX file exists at the path specified in the manifest and pointed to by spine toc
     }
 
     bool found = false;
@@ -903,25 +925,29 @@ void ImportEPUB::LocateOrCreateNCX(const QString &ncx_id_on_spine)
         ncx_href = m_NcxCandidates[ m_NCXId ];
         m_NCXFilePath = QFileInfo(m_OPFFilePath).absolutePath() % "/" % ncx_href;
         m_NCXFilePath = Utility::resolveRelativeSegmentsInFilePath(m_NCXFilePath, "/");
-        QString ncx_text = CleanSource::ProcessXML(Utility::ReadUnicodeTextFile(m_NCXFilePath),
-                                                   "application/x-dtbncx+xml");
-        QString bookpath = m_NCXFilePath.right(m_NCXFilePath.length() - m_ExtractedFolderPath.length() - 1);
-        NCXResource* resource = m_Book->GetFolderKeeper()->AddNCXToFolder(m_PackageVersion, bookpath);
-        resource->SetText(ncx_text);
-        resource->SaveToDisk(false);
-        if (m_FileInfoFromZip.contains(bookpath)) {
-            std::tuple<size_t, QString, QString> ainfo = m_FileInfoFromZip[bookpath];
-            resource->SetSavedSize(std::get<0>(ainfo));
-            resource->SetSavedCRC32(std::get<1>(ainfo));
-            resource->SetSavedDate(std::get<2>(ainfo));
-	    }
-        m_NCXNotInManifest = false;
-        load_warning = QObject::tr("The OPF file did not identify the NCX file correctly.") + "\n" + 
+        if (QFile::exists(m_NCXFilePath)) {
+            QString ncx_text = CleanSource::ProcessXML(Utility::ReadUnicodeTextFile(m_NCXFilePath),
+                                                       "application/x-dtbncx+xml");
+            QString bookpath = m_NCXFilePath.right(m_NCXFilePath.length() - m_ExtractedFolderPath.length() - 1);
+            NCXResource* resource = m_Book->GetFolderKeeper()->AddNCXToFolder(m_PackageVersion, bookpath);
+            resource->SetText(ncx_text);
+            resource->SaveToDisk(false);
+            if (m_FileInfoFromZip.contains(bookpath)) {
+                std::tuple<size_t, QString, QString> ainfo = m_FileInfoFromZip[bookpath];
+                resource->SetSavedSize(std::get<0>(ainfo));
+                resource->SetSavedCRC32(std::get<1>(ainfo));
+                resource->SetSavedDate(std::get<2>(ainfo));
+	        }
+            m_NCXNotInManifest = false;
+            load_warning = QObject::tr("The OPF file did not identify the NCX file correctly.") + "\n" + 
                                " - "  +  QObject::tr("Sigil has used the following file as the NCX:") + 
                                QString(" %1").arg(m_NcxCandidates[ m_NCXId ]);
 
-        AddLoadWarning(load_warning);
-        return;
+            AddLoadWarning(load_warning);
+            return;
+        }
+        // again file with ncx extension in manifest is missing
+        found = false;
     }
 
     // An NCX is only required in epub2 so punt here if epub3
@@ -960,6 +986,7 @@ void ImportEPUB::LocateOrCreateNCX(const QString &ncx_id_on_spine)
     // now add the NCX to our folder
     QString bookpath = m_NCXFilePath.right(m_NCXFilePath.length() - m_ExtractedFolderPath.length() - 1);
     NCXResource* nresource = m_Book->GetFolderKeeper()->AddNCXToFolder(m_PackageVersion, bookpath);
+    Q_UNUSED(nresource);
     if (!load_warning.isEmpty()) {
         AddLoadWarning(load_warning);
     }
@@ -998,19 +1025,11 @@ bool ImportEPUB::LoadFolderStructure()
     for (int i = 0; i < num_files; ++i) {
         QString id = keys.at(i);
         sync.addFuture(
-#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
-                       QtConcurrent::run(
-                           this,
-                           &ImportEPUB::LoadOneFile,
-                           m_Files.value(id),
-                           m_FileMimetypes.value(id))
-#else
                        QtConcurrent::run(
                            &ImportEPUB::LoadOneFile,
                            this,
                            m_Files.value(id),
                            m_FileMimetypes.value(id))
-#endif
                       );
     }
 
